from typing import Tuple

from pydantic import BaseModel, Field, ConfigDict
from typing_extensions import Literal

from GANDLF.optimizers import global_optimizer_dict

# takes the keys from global optimizer
OPTIMIZER_OPTIONS = Literal[tuple(global_optimizer_dict.keys())]


class SgdConfig(BaseModel):
    momentum: float = Field(default=0.99)
    weight_decay: float = Field(default=3e-05)
    dampening: float = Field(default=0)
    nesterov: bool = Field(default=True)


class AsgdConfig(BaseModel):
    alpha: float = Field(default=0.75)
    t0: float = Field(default=1e6)
    lambd: float = Field(default=1e-4)
    weight_decay: float = Field(default=3e-05)


class AdamConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    weight_decay: float = Field(default=0.00005)
    eps: float = Field(default=1e-8)
    amsgrad: bool = Field(default=False)


class AdamaxConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    weight_decay: float = Field(default=0.00005)
    eps: float = Field(default=1e-8)


class RpropConfig(BaseModel):
    etas: Tuple[float, float] = Field(default=(0.5, 1.2))
    step_sizes: Tuple[float, float] = Field(default=(1e-6, 50))


class AdadeltaConfig(BaseModel):
    rho: float = Field(default=0.9)
    eps: float = Field(default=1e-6)
    weight_decay: float = Field(default=3e-05)


class AdagradConfig(BaseModel):
    lr_decay: float = Field(default=0)
    eps: float = Field(default=1e-6)
    weight_decay: float = Field(default=3e-05)


class RmspropConfig(BaseModel):
    alpha: float = Field(default=0.99)
    eps: float = Field(default=1e-8)
    centered: bool = Field(default=False)
    momentum: float = Field(default=0)
    weight_decay: float = Field(default=3e-05)


class RadamConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    eps: float = Field(default=1e-8)
    weight_decay: float = Field(default=3e-05)
    foreach: bool = Field(default=None)


class NadamConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    eps: float = Field(default=1e-8)
    weight_decay: float = Field(default=3e-05)
    foreach: bool = Field(default=None)


class NovogradConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    eps: float = Field(default=1e-8)
    weight_decay: float = Field(default=3e-05)
    amsgrad: bool = Field(default=False)


class AdemamixConfig(BaseModel):
    pass


class LionConfig(BaseModel):
    betas: Tuple[float, float] = Field(default=(0.9, 0.999))
    weight_decay: float = Field(default=0.0)
    decoupled_weight_decay: bool = Field(default=False)


class AdoptConfig(BaseModel):
    pass


class OptimizerConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    type: OPTIMIZER_OPTIONS = Field(description="Type of optimizer to use")


optimizer_dict_config = {
    "sgd": SgdConfig,
    "asgd": AsgdConfig,
    "adam": AdamConfig,
    "adamw": AdamConfig,
    "adamax": AdamaxConfig,
    # "sparseadam": sparseadam,
    "rprop": RpropConfig,
    "adadelta": AdadeltaConfig,
    "adagrad": AdagradConfig,
    "rmsprop": RmspropConfig,
    "radam": RadamConfig,
    "novograd": NovogradConfig,
    "nadam": NadamConfig,
    "ademamix": AdemamixConfig,
    "lion": LionConfig,
    "adopt": AdoptConfig,
}



# -*- coding: utf-8 -*-
"""Add Conv Block for SDNet"""


def add_downsample_conv_block(
    Conv, BatchNorm, in_ch=1, out_ch=1, kernel_size=3, stride=2, dilate=1, last=False
):
    """
    Helper function
    """
    pad = dilate if not last else 0
    conv_1 = Conv(
        in_ch, out_ch, kernel_size, stride=stride, padding=pad, dilation=dilate
    )
    bn_1 = BatchNorm(out_ch)

    return [conv_1, bn_1]



from functools import partial
from typing import List
import torch

from torchio.transforms import Lambda


def axis_check(axis: List[int]) -> List[int]:
    """
    This function checks the axis for rotation.

    Args:
        axis (List[int]): The axes of rotation.

    Returns:
        List[int]: The affected axes.
    """
    if isinstance(axis, int):
        if axis == 0:
            axis = [1]
        else:
            axis = [axis]
    if 0 in axis:
        for count, _ in enumerate(axis):
            axis[count] += 1
    for sub_ax in axis:
        assert isinstance(
            sub_ax, int
        ), f"Axis must be an integer, but was provided as: {sub_ax}"
        assert sub_ax in [
            1,
            2,
            3,
        ], f"Axes must be in [1, 2, 3], but was provided as: {sub_ax}"

    relevant_axes = set([1, 2, 3])
    if relevant_axes == set(axis):
        affected_axes = list(relevant_axes)
    else:
        affected_axes = list(relevant_axes - set(axis))

    if len(affected_axes) == 1:
        affected_axes.append(affected_axes[0])
    return affected_axes


def tensor_rotate_90(input_image: torch.Tensor, axis: List[int]) -> torch.Tensor:
    """
    This function rotates an image by 90 degrees around the specified axis.

    Args:
        input_image (torch.Tensor): The input tensor.
        axis (List[int]): The axes of rotation.

    Returns:
        torch.Tensor: The rotated tensor.
    """
    # with 'axis' axis of rotation, rotate 90 degrees
    # tensor image is expected to be of shape (1, a, b, c)
    # if 0 is in axis, ensure it is not considered, since that is the batch dimension
    affected_axes = axis_check(axis)
    return torch.transpose(input_image, affected_axes[0], affected_axes[1]).flip(
        affected_axes[1]
    )


def tensor_rotate_180(input_image: torch.Tensor, axis: List[int]) -> torch.Tensor:
    """
    This function rotates an image by 180 degrees around the specified axis.

    Args:
        input_image (torch.Tensor): The input tensor.
        axis (List[int]): The axes of rotation.

    Returns:
        torch.Tensor: The rotated tensor.
    """
    # with 'axis' axis of rotation, rotate 180 degrees
    # tensor image is expected to be of shape (1, a, b, c)
    affected_axes = axis_check(axis)
    return input_image.flip(affected_axes[0]).flip(affected_axes[1])


def rotate_90(parameters: dict) -> Lambda:
    """
    This function rotates an image by 90 degrees around the specified axis.

    Args:
        parameters (dict): The parameters for the rotation.

    Returns:
        Lambda: The rotation function.
    """
    return Lambda(
        function=partial(tensor_rotate_90, axis=parameters["axis"]),
        p=parameters["probability"],
    )


def rotate_180(parameters: dict) -> Lambda:
    """
    This function rotates an image by 180 degrees around the specified axis.

    Args:
        parameters (dict): The parameters for the rotation.

    Returns:
        Lambda: The rotation function.
    """
    return Lambda(
        function=partial(tensor_rotate_180, axis=parameters["axis"]),
        p=parameters["probability"],
    )



import torch.nn as nn


class EncodingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        norm=nn.BatchNorm2d,
        norm_kwargs=None,
        act=nn.LeakyReLU,
        act_kwargs=None,
        dropout=nn.Dropout2d,
        dropout_kwargs=None,
        network_kwargs=None,
    ):
        """
        Encoding module for the UNet model.

        Args:
            input_channels (int): Number of input channels.
            output_channels (int): Number of output channels.
            conv (nn.Module): Convolution layer.
            conv_kwargs (dict): Convolution layer keyword arguments.
            norm (nn.Module): Normalization layer.
            norm_kwargs (dict): Normalization layer keyword arguments.
            act (nn.Module): Activation function layer.
            act_kwargs (dict): Activation function layer keyword arguments.
            dropout (nn.Module): Dropout layer.
            dropout_kwargs (dict): Dropout layer keyword arguments.
            residual (bool): Flag for using residual connection.
        """
        nn.Module.__init__(self)
        # Dev note : This should have been a super
        # super(EncodingModule, self).__init__()
        # but need to test it more

        # Set default arguments
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 3, "stride": 1, "padding": 1, "bias": True}
        if norm_kwargs is None:
            norm_kwargs = {
                "eps": 1e-5,
                "affine": True,
                "momentum": 0.1,
                "track_running_stats": True,
            }
        if act_kwargs is None:
            act_kwargs = {"negative_slope": 1e-2, "inplace": True}
        if dropout_kwargs is None:
            dropout_kwargs = {"p": 0.5, "inplace": False}
        if network_kwargs is None:
            network_kwargs = {"res": False}

        # Set attributes together
        self.residual = network_kwargs["res"]
        self.act = act(**act_kwargs)
        self.dropout = dropout(**dropout_kwargs) if dropout is not None else None

        # Define layers together
        self.in_0 = norm(output_channels, **norm_kwargs)
        self.in_1 = norm(output_channels, **norm_kwargs)

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)
        self.conv1 = conv(output_channels, output_channels, **conv_kwargs)

    def forward(self, x):
        """
        The forward function for encoding module.

        [input --> | --> in --> lrelu --> conv0 --> dropout --> in -|
                   |                                                |
        output <-- + <-------------------------- conv1 <-- lrelu <--|]

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        if self.residual:
            skip = x
        x = self.act(self.in_0(x))

        x = self.conv0(x)
        if self.dropout is not None:
            x = self.dropout(x)
        x = self.act(self.in_1(x))

        x = self.conv1(x)
        if self.residual:
            x = x + skip

        return x
